import {Gamesystem} from "./Gamesystem";
import {ProductState} from "./ProductState";
import {ProductTransition} from "./ProductTransition";
import {State} from "./State";
import {Transition} from "./Transition";
import {SimpleState} from "./SimpleState";
import {SimpleGamesystem} from "./SimpleGamesystem";
import {GameModel} from "../GameModel";
import {ProductStateTrainer} from "../../../../e2e/game-model/gamesystems/productGamesystems/ProductStateTrainer";

export class ProductGamesystem extends Gamesystem<ProductState, ProductTransition> {

  innerGamesystems: Gamesystem<State<any>, Transition<any>>[] = [];
  parentGamesystem: ProductGamesystem | undefined

  static constructFromSimpleGamesystem(simpleGamesystem: SimpleGamesystem, gameModel: GameModel) {
    const productGamesystem = new ProductGamesystem(simpleGamesystem.componentName);
    const parentGamesystem = simpleGamesystem.parentGamesystem;

    if(simpleGamesystem.states.length > 0) {
      simpleGamesystem.componentName += "(Child)";
      productGamesystem.addChildGamesystem(simpleGamesystem);
    }


    if(parentGamesystem != undefined) {
      parentGamesystem.removeChildGamesystem(simpleGamesystem);
      parentGamesystem.addChildGamesystem(productGamesystem);
    } else {
      gameModel.removeGamesystem(simpleGamesystem);
      gameModel.addGamesystem(productGamesystem);
    }

    return productGamesystem;
  }

  createState(innerStates: State<any>[]): ProductState | undefined {
    if(innerStates.length == this.innerGamesystems.length && this.findProductStateByInnerStates(innerStates)== undefined) {
      const productState = new ProductState(innerStates);
      this.states.push(productState);
      return productState;
    } else {
      return undefined;
    }
  }

  createTransition(startingState: ProductState, endingState: ProductState): ProductTransition | undefined {
    if(startingState == undefined || endingState == undefined || this.existsTransition(startingState, endingState)
            || !this.existsState(startingState) || !this.existsState(endingState) || startingState.equals(endingState)) {
      return undefined;
    }

    const productTransition = new ProductTransition(startingState, endingState);
    this.transitions.push(productTransition);
    return productTransition;
  }

  removeState(state: ProductState): boolean {
    return false;
  }

  removeInnerState(state: SimpleState) {

  }


  addChildGamesystem(gamesystem: Gamesystem<State<any>, Transition<any>>) {
    this.innerGamesystems.push(gamesystem);
  }

  private removeChildGamesystem(gamesystem: Gamesystem<State<any>, Transition<any>>) {
    this.innerGamesystems = this.innerGamesystems.filter(childSystem => childSystem != gamesystem);
  }

  findProductStateByInnerStates(innerStates: State<any>[]) {
    return this.states.find(productState => productState.equalInnerStates(innerStates));
  }

  findChildSystemByName(gamesystemName: string) {
    const gamesystemQueue: Gamesystem<State<any>, Transition<any>>[] = [];
    this.innerGamesystems.forEach(gamesystem => gamesystemQueue.push(gamesystem));

    while(gamesystemName.length > 0 ){
      const currentGamesystem = gamesystemQueue.shift();
      if(currentGamesystem!.componentName === gamesystemName) {
        return currentGamesystem;
      }
    }
    return undefined;
  }

  private existsTransition(startingState: ProductState, endingState: ProductState) {
    return this.transitions.find(transition =>
      transition.startingState.equals(startingState) && transition.endingState.equals(endingState)) != undefined;
  }

  private existsState(state: ProductState) {
    return this.states.find(s => s.equals(state)) != undefined;
  }
}
